
import os
import sys
import time
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional
import prov.model as prov
# import pwd
import getpass as gt
import warnings
from aenum import extend_enum
import subprocess

from yprov4ml.datamodel.artifact_data import ArtifactInfo
from yprov4ml.datamodel.attribute_type import LoggingItemKind
from yprov4ml.datamodel.metric_data import MetricInfo
from yprov4ml.datamodel.context import Context
from yprov4ml.datamodel.metric_type import MetricsType
from yprov4ml.datamodel.compressor_type import CompressorType, COMPRESSORS_FOR_ZARR
from yprov4ml.utils import funcs
from yprov4ml.utils.prov_utils import get_activity, create_activity
from yprov4ml.utils.funcs import get_global_rank, get_runtime_type

class Prov4MLData:
    def __init__(self) -> None:
        self.metrics: Dict[(str, Context), MetricInfo] = {}
        self.artifacts: Dict[(str, Context), ArtifactInfo] = {}

        self.PROV_SAVE_PATH = "prov_save_path"
        self.PROV_JSON_NAME = "test_experiment"
        self.EXPERIMENT_DIR = "test_experiment_dir"
        self.ARTIFACTS_DIR = "artifact_dir"
        self.METRIC_DIR = "metric_dir"
        self.USER_NAMESPACE = "user_namespace"
        self.PROV_PREFIX = "yProv4ML"
        self.RUN_ID = 0

        self.global_rank = None
        self.is_collecting = False

        self.save_metrics_after_n_logs = 100
        self.csv_separator = ","

    def start_run(
            self, 
            experiment_name: str, 
            prov_save_path: Optional[str] = None, 
            user_namespace: Optional[str] = None, 
            collect_all_processes: bool = False, 
            save_after_n_logs: int = 100, 
            rank: Optional[int] = None, 
            metrics_file_type: MetricsType = MetricsType.ZARR,
            csv_separator:str = ",", 
            use_compressor: Optional[CompressorType] = None,
        ) -> None:

        self.global_rank = funcs.get_global_rank() if rank is None else rank
        self.is_collecting = self.global_rank is None or int(self.global_rank) == 0 or collect_all_processes
        
        if not self.is_collecting: return

        self.save_metrics_after_n_logs = save_after_n_logs
        if prov_save_path: self.PROV_SAVE_PATH = prov_save_path
        if user_namespace: self.USER_NAMESPACE = user_namespace

        if use_compressor in COMPRESSORS_FOR_ZARR and metrics_file_type != MetricsType.ZARR: 
            warnings.warn(f">start_run(): use_compressor chosen is only compatible with MetricsType.ZARR, but saving type is {metrics_file_type}, the compressor chosen will have no effect")
        if metrics_file_type == MetricsType.ZARR and use_compressor != False and use_compressor not in COMPRESSORS_FOR_ZARR: 
            raise AttributeError(f">start_run(): use_compressor chosen is only compatible with MetricsType.ZARR")

        if metrics_file_type == MetricsType.ZARR and use_compressor:
            use_compressor = CompressorType.BLOSC_ZSTD
        elif metrics_file_type in [MetricsType.NETCDF, MetricsType.CSV] and use_compressor:
            use_compressor = CompressorType.ZIP
        if not use_compressor: 
            use_compressor = CompressorType.NONE

        # look at PROV dir how many experiments are there with the same name
        if not os.path.exists(self.PROV_SAVE_PATH):
            os.makedirs(self.PROV_SAVE_PATH, exist_ok=True)
            self.RUN_ID = 0
        else: 
            prev_exps = os.listdir(self.PROV_SAVE_PATH) 
            matching_files = [int(exp.split("_")[-1].split(".")[0]) for exp in prev_exps if funcs.prov4ml_experiment_matches(experiment_name, exp)]
            self.RUN_ID = max(matching_files)+1  if len(matching_files) > 0 else 0

        self.CLEAN_EXPERIMENT_NAME = experiment_name
        self.PROV_JSON_NAME = self.CLEAN_EXPERIMENT_NAME + f"_GR{self.global_rank}" if self.global_rank else experiment_name + f"_GR0"
        self.PROV_JSON_NAME = f"{self.PROV_JSON_NAME}_{self.RUN_ID}"

        self.EXPERIMENT_DIR = os.path.join(self.PROV_SAVE_PATH, f"{self.CLEAN_EXPERIMENT_NAME}_{self.RUN_ID}")
        self.ARTIFACTS_DIR = os.path.join(self.EXPERIMENT_DIR, f"artifacts_GR{self.global_rank}")
        self.METRIC_DIR = os.path.join(self.EXPERIMENT_DIR, f"metrics_GR{self.global_rank}")

        self.metrics_file_type = metrics_file_type
        self.use_compressor = use_compressor
        self.csv_separator = csv_separator

        self._init_root_context()

        # necessary when spawning threads, 
        # otherwise they get counted as different runs
        # TODO: find better approach
        time.sleep(1)
        os.makedirs(self.EXPERIMENT_DIR, exist_ok=True)
        os.makedirs(self.ARTIFACTS_DIR, exist_ok=True)
        os.makedirs(self.METRIC_DIR, exist_ok=True)


    def _add_ctx(self, rootContext, ctx):
        c = self.root_provenance_doc.activity("context:"+ str(ctx))
        c.wasInformedBy(rootContext)
        c.add_attributes({f'{self.PROV_PREFIX}:level':1})

    def _init_root_context(self): 
        self.root_provenance_doc = prov.ProvDocument()
        self.root_provenance_doc.add_namespace('context', 'context')
        self.root_provenance_doc.add_namespace(self.PROV_PREFIX, self.PROV_PREFIX)
        self.root_provenance_doc.set_default_namespace(self.PROV_JSON_NAME)
        # self.root_provenance_doc.set_default_namespace(self.USER_NAMESPACE)
        self.root_provenance_doc.add_namespace('prov','http://www.w3.org/ns/prov#')
        self.root_provenance_doc.add_namespace('xsd','http://www.w3.org/2000/10/XMLSchema#')
        self.root_provenance_doc.add_namespace('prov-ml', 'prov-ml')
        # self.provDoc.add_namespace(name,name)

        user_ag = self.root_provenance_doc.agent(f'{self.PROV_PREFIX}:{gt.getuser()}')
        rootContext = self.root_provenance_doc.activity("context:"+ self.PROV_JSON_NAME)
        rootContext.add_attributes({
            f'{self.PROV_PREFIX}:level':0, 
            f"{self.PROV_PREFIX}:provenance_path":self.PROV_SAVE_PATH,
            f"{self.PROV_PREFIX}:artifact_uri":self.ARTIFACTS_DIR,
            f"{self.PROV_PREFIX}:experiment_dir":self.EXPERIMENT_DIR,
            f"{self.PROV_PREFIX}:experiment_name":self.PROV_JSON_NAME,
            f"{self.PROV_PREFIX}:run_id":self.RUN_ID,
            f"{self.PROV_PREFIX}:python_version":str(sys.version), 
        })
        rootContext.wasAssociatedWith(user_ag)
        self._add_ctx(rootContext, Context.TRAINING)
        self._add_ctx(rootContext, Context.VALIDATION)
        self._add_ctx(rootContext, Context.TESTING)
        self._add_ctx(rootContext, Context.DATASETS)
        self._add_ctx(rootContext, Context.MODELS)

        global_rank = get_global_rank()
        runtime_type = get_runtime_type()
        if runtime_type == "slurm":
            node_rank = os.getenv("SLURM_NODEID", None)
            local_rank = os.getenv("SLURM_LOCALID", None) 
            rootContext.add_attributes({
                f"{self.PROV_PREFIX}:global_rank": global_rank,
                f"{self.PROV_PREFIX}:local_rank": local_rank,
                f"{self.PROV_PREFIX}:node_rank": node_rank,
            })
        elif runtime_type == "single_core":
            rootContext.add_attributes({
                f"{self.PROV_PREFIX}:global_rank": global_rank
            })


        
    def _log_input(self, path : str, context : Context, attributes : dict={}) -> prov.ProvEntity:
        entity = self.root_provenance_doc.entity(path, attributes)
        activity = get_activity(self.root_provenance_doc,"context:"+str(context))
        activity.used(entity)
        return entity
    
    def _log_output(self, path : str, context : Context, attributes : dict={}) -> prov.ProvEntity:
        entity= self.root_provenance_doc.entity(path, attributes)
        activity = get_activity(self.root_provenance_doc,"context:"+str(context))
        entity.wasGeneratedBy(activity)
        return entity
    
    def _get_git_revision_hash() -> str:
        return subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode('ascii').strip()

    def _get_git_remote_url() -> Optional[str]:
        try:
            remote_url = subprocess.check_output(['git', 'config', '--get', 'remote.origin.url'], stderr=subprocess.DEVNULL).strip().decode()
            return remote_url
        except subprocess.CalledProcessError:
            print("> get_git_remote_url() Repository not found")
            return None  # No remote found

    def add_source_code(self, path: str): 
        if path is None:
            repo = _get_git_remote_url()
            if repo is not None:
                commit_hash = _get_git_revision_hash()
                log_param("source_code", os.path.join(repo, commit_hash))
        else:
            p = Path(path)
            if p.is_file():
                self.add_artifact(p.name.replace(".py", ""), str(p), log_copy_in_prov_directory=True, is_model=False, is_input=True)
                self.add_parameter("source_code", os.path.join(self.ARTIFACTS_DIR, p.name))
            else:
                self.add_artifact("source_code", str(p), log_copy_in_prov_directory=True, is_model=False, is_input=True)
                self.add_parameter("source_code", os.path.join(self.ARTIFACTS_DIR, "source_code"))


    def add_context(self, context : str, is_subcontext_of: Optional[Context] = None):     
        is_subcontext_of = str(is_subcontext_of).split(".")[-1]

        extend_enum(Context, context, str(context))
        new_context = create_activity(self.root_provenance_doc,'context:'+ str(getattr(Context, context)))

        if is_subcontext_of is not None:
            if not hasattr(Context, is_subcontext_of): 
                raise Exception(f"{is_subcontext_of} not found as a valid Context")
            parent_context = get_activity(self.root_provenance_doc,'context:' + str(getattr(Context, is_subcontext_of)))
        else: 
            parent_context = get_activity(self.root_provenance_doc,'context:'+ str(Context.EXPERIMENT))

        level=list(parent_context.get_attribute(f'{self.PROV_PREFIX}:level'))[0]
        new_context.wasInformedBy(parent_context)
        new_context.add_attributes({f'{self.PROV_PREFIX}:level':level+1})

    def add_metric(
        self, 
        metric: str, 
        value: Any, 
        step: Optional[int] = None, 
        context: Optional[Any] = None, 
        source: Optional[LoggingItemKind] = None, 
    ) -> None:
        if not self.is_collecting: return

        if context is None: context = self.EXPERIMENT_NAME
        if step is None: step = 0

        if (metric, context) not in self.metrics:
            self.metrics[(metric, context)] = MetricInfo(metric, context, source=source, use_compressor=self.use_compressor)
        
        self.metrics[(metric, context)].add_metric(value, step, funcs.get_current_time_millis())

        total_metrics_values = self.metrics[(metric, context)].total_metric_values
        if total_metrics_values % self.save_metrics_after_n_logs == 0:
            self.save_metric_to_file(self.metrics[(metric, context)])

    def add_parameter(
            self, 
            parameter_name: str, 
            parameter_value: Any, 
            context : Optional[Context] = None, 
        ) -> None:
        """
        Adds a parameter to the provenance data.

        Parameters:
        -----------
        parameter : str
            The name of the parameter to add.
        value : Any
            The value of the parameter to add.

        Returns:
        --------
        None
        """
        if not self.is_collecting: return

        if context is None: 
            context = self.PROV_JSON_NAME

        current_activity = get_activity(self.root_provenance_doc,"context:"+ str(context))

        SUPPORTED_TYPES = [int, float, tuple, str, bool]
        if type(parameter_value) in SUPPORTED_TYPES: 
            current_activity.add_attributes({
                f"{self.PROV_PREFIX}:{parameter_name}": parameter_value
            })
        else: 
            current_activity.add_attributes({
                f"{self.PROV_PREFIX}:{parameter_name}": str({
                    "$": str(parameter_value), 
                    "type": type(parameter_value).__name__, 
                })
            })

    def add_artifact(
        self, 
        artifact_name: str, 
        artifact_path: str, 
        step: Optional[int] = None, 
        context: Optional[Any] = None,
        is_input : bool = False, 
        log_copy_in_prov_directory : bool = True, 
        is_model = False, 
    ) -> prov.ProvEntity:
        if not self.is_collecting: return

        if context is None: 
            context = self.PROV_JSON_NAME

        if not isinstance(artifact_path, str): 
            print(artifact_path)
            raise AttributeError(f">add_artifact({artifact_path}): the parameter \"artifact_path\" has to be a string")

        if log_copy_in_prov_directory: 
            try: 
                path = Path(artifact_path)

                # create original artefact
                original = self.add_artifact(f"original_{path.name}", str(path), log_copy_in_prov_directory=False, is_model=is_model, is_input=is_input)
                copied = self.add_artifact(f"{path.name}", os.path.join(self.ARTIFACTS_DIR, path.name), log_copy_in_prov_directory=False, is_model=is_model, is_input=True)
                copied.wasDerivedFrom(original)

                newart_path = os.path.join(self.ARTIFACTS_DIR, path.name)
                if path.is_file():
                    shutil.copy(path, newart_path)
                else:  
                    shutil.copytree(path, newart_path)
                artifact_path = newart_path

                return copied
            except: 
                Exception(f">add_artifact: log_copy_in_prov_directory was True but value is not a valid Path: {artifact_path}")

        self.artifacts[(f"{self.PROV_PREFIX}:{artifact_name}", context)] = ArtifactInfo(f"{self.PROV_PREFIX}:{artifact_name}", artifact_path, step, context=context, is_model=is_model)

        attributes = {
            f'{self.PROV_PREFIX}:label': artifact_name, 
            f'{self.PROV_PREFIX}:path': artifact_path,
        }

        if is_input: 
            attributes.setdefault(f'{self.PROV_PREFIX}:role','input')
            return self._log_input(f"{self.PROV_PREFIX}:{artifact_name}", context, attributes)
        else: 
            attributes.setdefault(f'{self.PROV_PREFIX}:role', 'output')
            return self._log_output(f"{self.PROV_PREFIX}:{artifact_name}", context, attributes)

    def get_artifacts(self) -> List[ArtifactInfo]:
        """
        Returns a list of all artifacts.

        Returns:
            List[ArtifactInfo]: A list of artifact information objects.
        """
        if not self.is_collecting: return

        return list(self.artifacts.values())
    
    def get_model_versions(self) -> List[ArtifactInfo]:
        """
        Returns a list of all model version artifacts.

        Returns:
            List[ArtifactInfo]: A list of model version artifact information objects.
        """
        if not self.is_collecting: return

        return [artifact for artifact in self.artifacts.values() if artifact.is_model_version]
    
    def get_final_model(self) -> Optional[ArtifactInfo]:
        """
        Returns the most recent model version artifact.

        Returns:
            Optional[ArtifactInfo]: The most recent model version artifact information object, or None if no model versions exist.
        """
        if not self.is_collecting: return

        model_versions = self.get_model_versions()
        if model_versions:
            return model_versions[-1]
        return None

    def save_metric_to_file(self, metric: MetricInfo) -> None:
        """
        Saves a metric to a temporary file.

        Parameters:
        --------
            metric (MetricInfo): The metric to save.
        
        Returns:
        --------
        None
        """
        if not self.is_collecting: return
        
        metric.save_to_file(self.METRIC_DIR, file_type=self.metrics_file_type, process=self.global_rank, csv_separator=self.csv_separator)

    def save_all_metrics(self) -> None:
        """
        Saves all tracked metrics to temporary files.

        Returns:
        --------
        None
        """
        if not self.is_collecting: return

        for metric in self.metrics.values():
            self.save_metric_to_file(metric)






import os
import pickle
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Union, Tuple
from typing_extensions import Literal

class LogMixin(metaclass=ABCMeta):
    @abstractmethod
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_kinds. Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """


class Logger(LogMixin, metaclass=ABCMeta):
    """Base class for logger

    Args:
        savedir (str, optional): filesystem location where logs are stored.
            Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            how often should the logger fulfill calls to the `log()`
            method:

            - When set to 'epoch', the logger logs only if ``batch_idx``
              is not passed to the ``log`` method.

            - When an integer
              is given, the logger logs if ``batch_idx`` is a multiple of
              ``log_freq``.

            - When set to ``'batch'``, the logger logs always.

            Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
    """
    #: Location on filesystem where to store data.
    savedir: str = None
    #: Supported logging 'kind's.
    supported_kinds: Tuple[str]
    #: Current worker global rank
    worker_rank: int

    _log_freq: Union[int, Literal['epoch', 'batch']]

    def __init__(
        self,
        savedir: str = 'mllogs',
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch',
        log_on_workers: Union[int, List[int]] = 0
    ) -> None:
        self.savedir = savedir
        self.log_freq = log_freq
        self.log_on_workers = log_on_workers

    @property
    def log_freq(self) -> Union[int, Literal['epoch', 'batch']]:
        """Get ``log_feq``, namely how often should the logger
        fulfill or ignore calls to the `log()` method."""
        return self._log_freq

    @log_freq.setter
    def log_freq(self, val: Union[int, Literal['epoch', 'batch']]):
        """Sanitize log_freq value."""
        if val in ['epoch', 'batch'] or (isinstance(val, int) and val > 0):
            self._log_freq = val
        else:
            raise ValueError(
                "Wrong value for 'log_freq'. Supported values are: "
                f"['epoch', 'batch'] or int > 0. Received: {val}"
            )

    @contextmanager
    def start_logging(self, rank: Optional[int] = None):
        """Start logging context.

        Args:
            rank (Optional[int]): global rank of current process,
                used in distributed environments. Defaults to None.

        Example:


        >>> with my_logger.start_logging():
        >>>     my_logger.log(123, 'value', kind='metric', step=0)


        """
        try:
            self.create_logger_context(rank=rank)
            yield
        finally:
            self.destroy_logger_context()

    @abstractmethod
    def create_logger_context(self, rank: Optional[int] = None) -> Any:
        """
        Initializes the logger context.

        Args:
            rank (Optional[int]): global rank of current process,
                used in distributed environments. Defaults to None.
        """

    @abstractmethod
    def destroy_logger_context(self) -> None:
        """Destroy logger."""

    @abstractmethod
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """

    def serialize(self, obj: Any, identifier: str) -> str:
        """Serializes object to disk and returns its path.

        Args:
            obj (Any): item to save.
            identifier (str): identifier of the item to log (expected to be a
                path under ``self.savedir``).

        Returns:
            str: local path of the serialized object to be logged.
        """
        itm_path = os.path.join(self.savedir, identifier)
        with open(itm_path, 'wb') as itm_file:
            pickle.dump(obj, itm_file)

    def should_log(
        self,
        batch_idx: Optional[int] = None
    ) -> bool:
        """Determines whether the logger should fulfill or ignore calls to the
        `log()` method, depending on the ``log_freq`` property:

        - When ``log_freq`` is set to 'epoch', the logger logs only if
          ``batch_idx`` is not passed to the ``log`` method.

        - When ``log_freq`` is an integer
          is given, the logger logs if ``batch_idx`` is a multiple of
          ``log_freq``.

        - When ``log_freq`` is set to ``'batch'``, the logger logs always.

        It also takes into account whether logging on the current worker
        rank is allowed by ``self.log_on_workers``.

        Args:
            batch_idx (Optional[int]): the dataloader batch idx, if available.
                Defaults to None.

        Returns:
            bool: True if the logger should log, False otherwise.
        """
        # Check worker's global rank
        worker_ok = (
            self.worker_rank is None or
            (isinstance(self.log_on_workers, int) and (
                self.log_on_workers == -1 or
                self.log_on_workers == self.worker_rank
            )
            )
            or
            (isinstance(self.log_on_workers, list)
             and self.worker_rank in self.log_on_workers)
        )
        if not worker_ok:
            return False

        # Check batch ID
        if batch_idx is not None:
            if isinstance(self.log_freq, int):
                if batch_idx % self.log_freq == 0:
                    return True
                return False
            if self.log_freq == 'batch':
                return True
            return False
        return True



import os
from typing import Any, Dict, List
from typing import Optional
import zarr
import netCDF4 as nc
import zarr.codecs

from yprov4ml.datamodel.attribute_type import LoggingItemKind
from yprov4ml.datamodel.compressor_type import CompressorType, compressor_to_type
from yprov4ml.datamodel.metric_type import MetricsType, get_file_type

ZARR_CHUNK_SIZE = 1000

class MetricInfo:
    """
    A class to store information about a specific metric.

    Attributes:
    -----------
    name : str
        The name of the metric.
    context : Any
        The context in which the metric is recorded.
    source : LoggingItemKind
        The source of the logging item.
    total_metric_values : int
        The total number of metric values recorded.
    epochDataList : dict
        A dictionary mapping epoch numbers to lists of metric values recorded in those epochs.

    Methods:
    --------
    __init__(name: str, context: Any, source=LoggingItemKind) -> None
        Initializes the MetricInfo class with the given name, context, and source.
    add_metric(value: Any, epoch: int, timestamp : int) -> None
        Adds a metric value for a specific epoch to the MetricInfo object.
    save_to_file(path : str, process : Optional[int] = None) -> None
        Saves the metric information to a file.
    """
    def __init__(self, 
                 name: str, 
                 context: Any, 
                 source=LoggingItemKind, 
                 use_compressor : Optional[CompressorType] = None
                 ) -> None:
        """
        Initializes the MetricInfo class with the given name, context, and source.

        Parameters:
        -----------
        name : str
            The name of the metric.
        context : Any
            The context in which the metric is recorded.
        source : LoggingItemKind
            The source of the logging item.

        Returns:
        --------
        None
        """
        self.name = name
        self.context = context
        self.source = source
        self.total_metric_values = 0
        self.use_compressor = use_compressor
        self.epochDataList: Dict[int, List[Any]] = {}

    def add_metric(self, value: Any, epoch: int, timestamp : int) -> None:
        """
        Adds a metric value for a specific epoch to the MetricInfo object.

        Parameters:
        -----------
        value : Any
            The value of the metric to be added.
        epoch : int
            The epoch number in which the metric value is recorded.
        timestamp : int
            The timestamp when the metric value was recorded.

        Returns:
        --------
        None
        """
        if epoch not in self.epochDataList:
            self.epochDataList[epoch] = []

        self.epochDataList[epoch].append((value, timestamp))
        self.total_metric_values += 1


    def save_to_file(
            self, 
            path: str, 
            file_type: MetricsType,
            csv_separator = ",", 
            process: Optional[int] = None, 
        ) -> None:
        """
        Saves the metric information to a file.

        Parameters
        ----------
        path : str
            The directory path where the file will be saved.
        file_type : str
            The type of file to be saved.
        process : Optional[int], optional
            The process identifier to be included in the filename. If not provided, 
            the filename will not include a process identifier.

        Returns
        -------
        None
        """
        process = process if process is not None else 0
        file = os.path.join(path, f"{self.name}_{self.context}_GR{process}")

        ft = file + get_file_type(file_type)
        if file_type == MetricsType.ZARR:
            self.save_to_zarr(ft)
        elif file_type == MetricsType.CSV:
            self.save_to_txt(ft, csv_separator)
        elif file_type == MetricsType.NETCDF:
            self.save_to_netCDF(ft)
        else:
            raise ValueError(f"Unsupported file type: {file_type}")

        self.epochDataList = {}


    def save_to_netCDF(
            self,
            netcdf_file: str
        ) -> None:
        """
        Saves the metric information in a netCDF file.

        Parameters
        ----------
        netcdf_file : str
            The path to the netCDF file where the metric information will be saved.

        Returns
        -------
        None
        """
        if os.path.exists(netcdf_file):
            dataset = nc.Dataset(netcdf_file, mode='a', format='NETCDF4')
        else:
            dataset = nc.Dataset(netcdf_file, mode='w', format='NETCDF4')
            dataset._name = self.name
            dataset._context = str(self.context)
            dataset._source = str(self.source)

            compression = 'zlib' if self.use_compressor else None
            dataset.createDimension('time', None)
            dataset.createVariable('epochs', 'i4', ('time',), compression)
            dataset.createVariable('values', 'f4', ('time',), compression)
            dataset.createVariable('timestamps', 'i8', ('time',), compression)

        epochs = []
        values = []
        timestamps = []

        for epoch, items in self.epochDataList.items():
            for value, timestamp in items:
                epochs.append(epoch)
                values.append(value)
                timestamps.append(timestamp)

        current_size = dataset.dimensions['time'].size
        new_size = current_size + len(epochs)

        dataset.variables['epochs'][current_size:new_size] = epochs
        dataset.variables['values'][current_size:new_size] = values
        dataset.variables['timestamps'][current_size:new_size] = timestamps

        dataset.close()

    def save_to_zarr(
            self,
            zarr_file: str
        ) -> None:
        """
        Saves the metric information in a zarr file.

        Parameters
        ----------
        zarr_file : str
            The path to the zarr file where the metric information will be saved.

        Returns
        -------
        None
        """
        if os.path.exists(zarr_file):
            dataset = zarr.open(zarr_file, mode='a')
        else:
            dataset = zarr.open(zarr_file, mode='w')

            # Metadata
            dataset.attrs['name'] = self.name
            dataset.attrs['context'] = str(self.context)
            dataset.attrs['source'] = str(self.source)

        epochs = []
        values = []
        timestamps = []

        for epoch, items in self.epochDataList.items():
            for value, timestamp in items:
                epochs.append(epoch)
                values.append(value)
                timestamps.append(timestamp)

        if 'epochs' not in dataset:
            dataset.create_array('epochs', shape=(0,), chunks=(ZARR_CHUNK_SIZE,), dtype='i4', compressors=compressor_to_type(self.use_compressor))
            dataset.create_array('values', shape=(0,), chunks=(ZARR_CHUNK_SIZE,), dtype='f4', compressors=compressor_to_type(self.use_compressor))
            dataset.create_array('timestamps', shape=(0,), chunks=(ZARR_CHUNK_SIZE,), dtype='i8', compressors=compressor_to_type(self.use_compressor))

        dataset['epochs'].append(epochs)
        dataset['values'].append(values)
        dataset['timestamps'].append(timestamps)

        dataset.store.close()

    def save_to_txt(
            self,
            txt_file: str, 
            csv_separator : str = ",", 
        ) -> None:
        """
        Saves the metric information in a text file.

        Parameters
        ----------
        txt_file : str
            The path to the text file where the metric information will be saved.

        Returns
        -------
        None
        """
        file_exists = os.path.exists(txt_file)

        with open(txt_file, "a") as f:
            if not file_exists:
                f.write(f"{self.name}{csv_separator}{self.context}{csv_separator}{self.source}\n")
            for epoch, values in self.epochDataList.items():
                for value, timestamp in values:
                    f.write(f"{epoch}{csv_separator}{value}{csv_separator}{timestamp}\n")



import os
from typing import Any, Dict, Optional, Union
from lightning.pytorch.loggers.logger import Logger
from typing_extensions import override
from argparse import Namespace
from torch import Tensor

from yprov4ml.logging_aux import log_param, log_metric
from yprov4ml.datamodel.context import Context

class ProvMLLogger(Logger):
    def __init__(
        self,
        name: Optional[str] = "lightning_logs",
        version: Optional[Union[int, str]] = None,
        prefix: str = "",
        flush_logs_every_n_steps: int = 100,
    ) -> None:
        """
        Initializes a ProvMLLogger instance.

        Parameters:
            name (Optional[str]): The name of the experiment. Defaults to "lightning_logs".
            version (Optional[Union[int, str]]): The version of the experiment. Defaults to None.
            prefix (str): The prefix for the experiment. Defaults to an empty string.
            flush_logs_every_n_steps (int): The number of steps after which logs should be flushed. Defaults to 100.
        """
        super().__init__()
        self._name = name or ""
        self._version = version
        self._prefix = prefix
        self._experiment = None
        self._flush_logs_every_n_steps = flush_logs_every_n_steps

    @property
    @override
    def root_dir(self) -> str:
        return os.path.join(self.save_dir, self.name)

    @property
    @override
    def log_dir(self) -> str:
        version = self.version if isinstance(self.version, str) else f"version_{self.version}"
        return os.path.join(self.root_dir, version)

    @property 
    @override
    def name(self) -> str:
        return self._name
    
    @property 
    @override
    def version(self) -> Optional[Union[int, str]]:
        return self._version
    
    @override
    def log_metrics(self, metrics: Dict[str, Union[Tensor, float]], step) -> None:

        for m, v in metrics.items(): 
            context = Context.TRAINING
            if "evaluation" in m or "test" in m: 
                context = Context.TESTING
            elif "validation" in m: 
                context = Context.VALIDATION
            log_metric(m, v, context=context)
    
    @override
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:
        for key, value in params.items():
            log_param(key, value)




import os
import argparse
from typing import Optional

def main(dot_file : str, out_file : Optional[str]): 
    if not dot_file.endswith(".dot"): 
        dot_file += ".dot"

    if out_file is None: 
        out_file = dot_file.replace(".dot", ".svg")
    if not out_file.endswith(".svg"): 
        out_file += ".svg"

    path_svg = os.path.join(".", out_file)
    os.system(f"dot -Tsvg {dot_file} > {path_svg}")
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Convert a DOT file to a SVG file')
    parser.add_argument('--dot', type=str, help='The DOT file to convert')
    parser.add_argument('--output', type=str, help='The output SVG file', default=None)
    args = parser.parse_args()

    main(args.dot, args.output)



from typing import Any, Callable, Tuple
from codecarbon import EmissionsTracker

def carbon_tracked_function(f: Callable, *args, **kwargs) -> Tuple[Any, Any]:
    """
    Tracks carbon emissions for a given function call.
    
    Args:
        f (Callable): The function to be executed and carbon emissions tracked.
        *args: Positional arguments to be passed to the function.
        **kwargs: Keyword arguments to be passed to the function.
    
    Returns:
        Tuple[Any, Any]: A tuple containing the result of the function call and the total emissions tracked.
    """
    TRACKER.start()
    result = f(*args, **kwargs)
    _ = TRACKER.stop()
    total_emissions = TRACKER._prepare_emissions_data()
    return result, total_emissions

def _carbon_init() -> None:
    """Initializes the carbon emissions tracker."""
    global TRACKER
    TRACKER = EmissionsTracker(
        save_to_file=False,
        save_to_api=False,
        save_to_logger=False, 
        log_level="error",
    ) #carbon emission tracker, don't save anywhere, just get the emissions value to log with prov4ml
    TRACKER.start()

def stop_carbon_tracked_block() -> Any:
    """
    Stops the tracking of carbon emissions for a code block and returns the total emissions tracked.
    
    Returns:
        Any: The total emissions tracked.
    """
    #_ = TRACKER.stop()
    total_emissions = TRACKER._prepare_emissions_data()
    return total_emissions




from typing import Any, Dict, Optional, Union
from typing_extensions import override
from typing import List, Tuple, Literal

from yprov4ml.logging_aux import *
from yprov4ml.loggers.itwinai_logger import Logger
from yprov4ml.yprov4ml import *

class ProvMLItwinAILogger(Logger):
    """
    Abstraction around Prov4ML logger.

    Args:
        prov_user_namespace (str, optional): location to where provenance
            files will be uploaded. Defaults to "www.example.org".
        experiment_name (str, optional): experiment name.
            Defaults to "experiment_name".
        provenance_save_dir (str, optional): path where to store provenance
            files and logs. Defaults to "prov".
        save_after_n_logs (Optional[int], optional): how often to save
            logs to disk from main memory. Defaults to 100.
        create_graph (Optional[bool], optional): whether to create a
            provenance graph. Defaults to True.
        create_svg (Optional[bool], optional): whether to create an SVG
            representation of the provenance graph. Defaults to True.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        log_on_workers (Optional[Union[int, List[int]]]): if -1, log on all
            workers; if int log on worker with rank equal to log_on_workers;
            if List[int], log on workers which rank is in the list.
            Defaults to 0 (the global rank of the main worker).
    """

    #: Supported kinds in the ``log`` method
    supported_kinds: Tuple[str] = (
        'metric', 'flops_pb', 'flops_pe', 'system', 'carbon',
        'execution_time', 'model', 'best_model',
        'torch')

    def __init__(
        self,
        prov_user_namespace="www.example.org",
        experiment_name="experiment_name",
        provenance_save_dir="mllogs",
        save_after_n_logs: Optional[int] = 100,
        create_graph: Optional[bool] = True,
        create_svg: Optional[bool] = True,
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch',
        log_on_workers: Union[int, List[int]] = 0
    ) -> None:
        super().__init__(
            savedir=provenance_save_dir,
            log_freq=log_freq,
            log_on_workers=log_on_workers
        )
        self.name = experiment_name
        self.version = None
        self.prov_user_namespace = prov_user_namespace
        self.provenance_save_dir = provenance_save_dir
        self.save_after_n_logs = save_after_n_logs
        self.create_graph = create_graph
        self.create_svg = create_svg

    @override
    def create_logger_context(self, rank: Optional[int] = None):
        """
        Initializes the logger context.

        Args:
            rank (Optional[int]): global rank of current process,
                used in distributed environments. Defaults to None.
        """
        self.worker_rank = rank

        if not self.should_log():
            return

        start_run(
            prov_user_namespace=self.prov_user_namespace,
            experiment_name=self.name,
            provenance_save_dir=self.provenance_save_dir,
            save_after_n_logs=self.save_after_n_logs,
            # This class will control which workers can log
            collect_all_processes=True,
            rank=rank
        )

    @override
    def destroy_logger_context(self):
        """
        Destroys the logger context.
        """
        if not self.should_log():
            return

        end_run(
            create_graph=self.create_graph,
            create_svg=self.create_svg
        )

    @override
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        if not self.should_log():
            return

        # Save hyperparams
        for param_name, val in params.items():
            log_param(param_name, val)

    @override
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        context: Optional[str] = 'training',
        **kwargs
    ) -> None:
        """Logs with Prov4ML.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_kinds``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """

        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == "metric":
            log_metric(key=identifier, value=item,
                               context=context, step=step)
        elif kind == "flops_pb":
            model, batch = item
            log_flops_per_batch(
                identifier, model=model,
                batch=batch, context=context, step=step)
        elif kind == "flops_pe":
            model, dataset = item
            log_flops_per_epoch(
                identifier, model=model,
                dataset=dataset, context=context, step=step)
        elif kind == "system":
            log_system_metrics(context=context, step=step)
        elif kind == "carbon":
            log_carbon_metrics(context=context, step=step)
        elif kind == "execution_time":
            log_current_execution_time(
                label=identifier, context=context, step=step)
        elif kind == 'model':
            save_model_version(
                model=item, model_name=identifier, context=context, step=step)
        elif kind == 'best_model':
            log_model(model=item, model_name=identifier, log_model_info=True, log_model_layers = False, is_input = False)
        elif kind == 'torch':
            from torch.utils.data import DataLoader
            if isinstance(item, DataLoader):
                log_dataset(dataset=item, dataset_label=identifier)
            else:
                log_param(key=identifier, value=item)




import os
import torch
from typing import Optional
import time

def _safe_get_model_attr(dic, node, attr, attr_label=None):
    if attr_label is None: 
        attr_label = str(attr) 

    try: 
        if attr == "type": 
            dic[attr_label] = str(type(node))
        elif attr == "weight.dtype": 
            dic[attr_label] = str(node.weight.dtype)
        else: 
            dic[attr_label] = str(getattr(node, attr))
    except AttributeError: 
        pass

def prov4ml_experiment_matches(
        experiment_name : str,
        exp_folder : str
    ) -> bool:
    """
    Check if the experiment name matches the experiment name in the provenance data.

    Parameters:
    -----------
    experiment_name : str
        The name of the experiment.
    exp_folder : str
        The name of the experiment folder.

    Returns:
    --------
    bool
        True if the experiment name matches the experiment name in the provenance data, False otherwise
    """
    exp_folder = "_".join(exp_folder.split("_")[:-1])
    return experiment_name == exp_folder

def get_current_time_millis() -> int:
    """
    Get the current time in milliseconds.

    Returns:
    --------
    int
        The current time in milliseconds.
    """
    return int(round(time.time() * 1000))

def get_global_rank() -> Optional[int]:
    """
    Retrieves the global rank of the current process in a distributed computing environment.

    This function checks if a distributed computing environment is available and initialized, and
    returns the global rank of the current process accordingly. The function supports two types of
    distributed environments:

    1. **PyTorch Distributed**: If PyTorch's distributed package is available and initialized,
       it retrieves the global rank of the process using `torch.distributed.get_rank()`.

    2. **SLURM**: If the environment variable `SLURM_PROCID` is present, it assumes that the process is
       managed by SLURM and retrieves the local rank from this environment variable.

    If neither distributed environment is detected, the function returns `None`.

    Returns:
    --------
    Optional[int]
        The global rank of the process if a distributed environment is detected, otherwise `None`.

    Examples:
    ---------
    >>> get_global_rank()
    0
    >>> get_global_rank()
    None
    """
    # if on torch.distributed, return the rank
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    
    # if on slurm, return the local rank
    if "SLURM_PROCID" in os.environ:
        return int(os.getenv("SLURM_PROCID", None))
    
    return 0

def get_runtime_type(): 
    """
    Get the runtime type.

    Returns:
    --------
    str
        The runtime type.

    Examples:
    ---------
    >>> get_runtime_type()
    "single_core"
    """
    if torch.distributed.is_available() and torch.distributed.is_initialized():
        return "distributed"
    return "single_core"



import sys
from enum import Enum

class LoggingItemKind(Enum): 
    """
    An enumeration representing different kinds of logging items used in provenance data collection.

    Attributes:
    -----------
    METRIC : str
        Represents a custom metric.
    FLOPS_PER_BATCH : str
        Represents the FLOPS (floating-point operations per second) calculated per batch.
    FLOPS_PER_EPOCH : str
        Represents the FLOPS calculated per epoch.
    SYSTEM_METRIC : str
        Represents system metrics related to hardware and system performance.
    CARBON_METRIC : str
        Represents carbon emission metrics.
    EXECUTION_TIME : str
        Represents the execution time of code segments.
    MODEL_VERSION : str
        Represents the version of the machine learning model.
    FINAL_MODEL_VERSION : str
        Represents the final version of the machine learning model.
    PARAMETER : str
        Represents parameters used in the experiment.

    Notes:
    ------
    - This enumeration helps categorize different types of logging items for better organization and management
      of provenance data.
    - Each item corresponds to a specific aspect of logging or metrics that might be tracked in an experiment.
    """
    METRIC = 'metric'
    FLOPS_PER_BATCH = 'flops_pb'
    FLOPS_PER_EPOCH = 'flops_pe'
    SYSTEM_METRIC = 'system'
    CARBON_METRIC = 'carbon'
    EXECUTION_TIME = 'execution_time'
    MODEL_VERSION = 'model_version'
    FINAL_MODEL_VERSION = 'model_version_final'
    PARAMETER = 'param'

def get_source_from_kind(kind: LoggingItemKind) -> str:
    """
    Returns the source string based on the logging item kind.

    Parameters:
    -----------
    kind : LoggingItemKind
        The type of logging item which determines the source.

    Returns:
    --------
    str
        The source string associated with the provided logging item kind.
    """
    if kind == LoggingItemKind.METRIC or kind == None:
        return 'custom_metric'
    elif kind == LoggingItemKind.FLOPS_PER_BATCH or kind == LoggingItemKind.FLOPS_PER_EPOCH:
        return 'fvcore.nn.FlopCountAnalysis'
    elif kind == LoggingItemKind.SYSTEM_METRIC:
        if sys.platform != 'darwin':
            return 'pyamdgpuinfo'
        else: 
            return "apple_gpu"            
    elif kind == LoggingItemKind.CARBON_METRIC:
        return 'codecarbon'
    elif kind == LoggingItemKind.EXECUTION_TIME:
        return 'std.time'
    elif kind == LoggingItemKind.MODEL_VERSION or kind == LoggingItemKind.FINAL_MODEL_VERSION:
        return 'torch'
    else:
        return ""





import psutil
import torch
import sys
import warnings

from yprov4ml.constants import VERBOSE

if sys.platform != 'darwin':
    import GPUtil
    import gpustat

    if torch.cuda.device_count() == 0:
        if VERBOSE:
            warnings.warn("No GPU found")
    elif "AMD" in torch.cuda.get_device_name(0): 
        import pyamdgpuinfo
    elif "NVIDIA" in torch.cuda.get_device_name(0): 
        from nvitop import Device
else: 
    import apple_gpu

def get_cpu_usage() -> float:
    """
    Returns the current CPU usage percentage.
    
    Returns:
        float: The CPU usage percentage.
    """
    return psutil.cpu_percent()

def get_memory_usage() -> float:
    """
    Returns the current memory usage percentage.
    
    Returns:
        float: The memory usage percentage.
    """
    return psutil.virtual_memory().percent

def get_disk_usage() -> float:
    """
    Returns the current disk usage percentage.
    
    Returns:
        float: The disk usage percentage.
    """
    return psutil.disk_usage('/').percent

def get_gpu_memory_usage() -> float:
    """
    Returns the current GPU memory usage percentage, if GPU is available.
    
    Returns:
        float: The GPU memory usage percentage.
    """    
    if sys.platform != 'darwin':
        if torch.cuda.is_available():
            return torch.cuda.memory_allocated() / torch.cuda.memory_reserved()
        else: 
            try: 
                if "NVIDIA" in torch.cuda.get_device_name(0):
                    devices = Device.all()
                    if len(devices) == 0:  
                        return 0.0
                    device = devices[0] 
                    return device.memory_used / device.memory_total
            except: 
                return 0.0
            return 0.0
    else: 
        return get_gpu_metric_apple('memory')

    
def get_gpu_power_usage() -> float:
    """
    Returns the current GPU power usage percentage, if GPU is available.
    
    Returns:
        float: The GPU power usage percentage.
    """
    if sys.platform != 'darwin':
        if torch.cuda.device_count() == 0:
            return 0.0

        gpu_power = 0.0
        if torch.cuda.is_available():
            gpu_power = get_gpu_metric_gputil('power')
        try: 
            if not gpu_power and "AMD" in torch.cuda.get_device_name(0):
                gpu_power = get_gpu_metric_amd('power')
        except: 
            return 0.0
    else:
        gpu_power = get_gpu_metric_apple('power')

    return gpu_power
    
def get_gpu_temperature() -> float:
    """
    Returns the current GPU temperature, if GPU is available.
    
    Returns:
        float: The GPU temperature.
    """
    if sys.platform != 'darwin':
        if torch.cuda.device_count() == 0:
            return 0.0

        gpu_temperature = 0.0
        if torch.cuda.is_available():
            gpu_temperature = get_gpu_metric_gputil('temperature')
        
        try: 
            if not gpu_temperature and "NVIDIA" in torch.cuda.get_device_name(0): 
                gpu_utilization = get_gpu_metric_nvidia('temperature')
            if not gpu_temperature and "AMD" in torch.cuda.get_device_name(0): 
                gpu_temperature = get_gpu_metric_amd('temperature')
        except: 
            gpu_temperature = 0.0

    else:
        gpu_temperature = get_gpu_metric_apple('temperature')

    return gpu_temperature

def get_gpu_usage() -> float:
    """
    Returns the current GPU usage percentage, if GPU is available.
    
    Returns:
        float: The GPU usage percentage.
    """
    if sys.platform != 'darwin':
        if torch.cuda.device_count() == 0:
            return 0.0
        
        gpu_utilization = 0.0
        if torch.cuda.is_available():
            gpu_utilization = get_gpu_metric_gputil('utilization')

        try: 
            if not gpu_utilization and "NVIDIA" in torch.cuda.get_device_name(0):
                gpu_utilization = get_gpu_metric_nvidia('utilization')
            if not gpu_utilization and "AMD" in torch.cuda.get_device_name(0):
                gpu_utilization = get_gpu_metric_amd('utilization')
        except: 
            gpu_utilization = 0.0
    else:
        gpu_utilization = get_gpu_metric_apple('utilization')

    return gpu_utilization

def get_gpu_metric_amd(metric): 
    try: 
        first_gpu = pyamdgpuinfo.get_gpu(0)
        if metric == 'power':
            m = first_gpu.query_power()
        elif metric == 'temperature':
            m = first_gpu.query_temperature()
        elif metric == 'utilization':
            m = first_gpu.query_utilization()

        return m
    except:
        if VERBOSE:
            warnings.warn(f"Could not get metric: {metric}")
        return 0.0

def get_gpu_metric_nvidia(metric):

    devices = Device.all()
    if len(devices) == 0:  
        return None
    device = devices[0] 

    if metric == 'temperature':
        return device.temperature()
    elif metric == "utilization": 
        return device.gpu_utilization()
    elif metric == 'fan_speed': 
        return device.fan_speed()
    elif metric == 'memory_total':
        return device.memory_total_human()
    else: 
        if VERBOSE:
            warnings.warn(f"Could not get metric: {metric}")
        return 0.0

def get_gpu_metric_gputil(metric):
    current_gpu = torch.cuda.current_device()
    gpus = GPUtil.getGPUs()
    if current_gpu < len(gpus):
        if metric == 'temperature':
            return gpus[current_gpu].temperature
        elif metric == "utilization": 
            return gpus[current_gpu].load
        else: 
            return 0.0
    else:
        if VERBOSE:
            warnings.warn(f"Could not get metric: {metric}")
        return 0.0

def get_gpu_metric_apple(metric):
    statistics = apple_gpu.accelerator_performance_statistics()
    if metric == 'power':
        return 0.0
    elif metric == 'temperature':
        return 0.0
    elif metric == 'utilization':
        return statistics['Device Utilization %']
    elif metric == 'memory':
        return statistics['Alloc system memory']
    else: 
        if VERBOSE:
            warnings.warn(f"Could not get metric: {metric}")
        return 0.0



import os
import sys
import prov
import prov.model as prov
from rocrate.rocrate import ROCrate
from pathlib import Path

from yprov4ml.constants import PROV4ML_DATA
from yprov4ml.datamodel.metric_type import get_file_type

def create_prov_document() -> prov.ProvDocument:
    
    doc = PROV4ML_DATA.root_provenance_doc
    # run_activity = get_activity(doc, "context:" + PROV4ML_DATA.EXPERIMENT_NAME)
    file_type = get_file_type(PROV4ML_DATA.metrics_file_type)

    for (name, ctx) in PROV4ML_DATA.metrics.keys():
        metric_file_path = os.path.join(PROV4ML_DATA.METRIC_DIR, name + "_" + str(ctx) + f"_GR{PROV4ML_DATA.global_rank}" + file_type)
        s = PROV4ML_DATA.metrics[(name, ctx)].source
        e = PROV4ML_DATA.add_artifact(name,metric_file_path,0,ctx, is_input=False, log_copy_in_prov_directory=False)
        
        e.add_attributes({
            f'{PROV4ML_DATA.PROV_PREFIX}:context': str(ctx),
            f'{PROV4ML_DATA.PROV_PREFIX}:source': str(s)
        })

    return doc

def get_properties_from_file(file : str):
    if file.endswith(".dot"): 
        return {
            "name": "pygraphviz provenance graph file",
            "encodingFormat": "application/dot"
        }
    elif file.endswith(".csv"): 
        return {
            "name": "metric",
            "encodingFormat": "text/csv"
        }
    elif file.endswith(".svg"): 
        return {
            "name": "pygraphviz svg provenance graph file",
            "encodingFormat": "image/svg+xml"
        }
    elif file.endswith(".json") and "/" not in file: 
        return {
            "name": "provenance JSON file",
            "encodingFormat": "text/json"
        }
    elif file.endswith(".json") and "/" in file: 
        return {
            "name": "JSON property description",
            "encodingFormat": "text/json"
        }
    elif file.endswith(".pt") or file.endswith(".pth"): 
        return {
            "name": "pytorch model checkpoint",
            "encodingFormat": "application/octet-stream"
        }
    elif file.endswith(".py"): 
        return {
            "name": "python source file",
            "encodingFormat": "text/plain"
        }
    else: 
        return {
            "name": file,
            "encodingFormat": f"{file.split('.')[-1]}",
        }

def create_rocrate_in_dir(directory): 
    crate = ROCrate()

    for (d, _, fs) in os.walk(directory): 
        for f in fs: 
            file_path = os.path.join(d, f)
            if Path(file_path).exists():
                property = get_properties_from_file(file_path)
                property["@type"] = "File" 
                property["@id"] = file_path
                crate.add_file(file_path, dest_path=file_path, properties=property)

    # crate.write("exp_crate")
    crate.write_zip(f"{directory}.zip")


from fvcore.nn import FlopCountAnalysis
from typing import Any

def _init_flops_counters() -> None:
    """Initializes the global FLOPs counters."""
    global FLOPS_PER_BATCH_COUNTER
    global FLOPS_PER_EPOCH_COUNTER
    FLOPS_PER_BATCH_COUNTER = 0
    FLOPS_PER_EPOCH_COUNTER = 0

def get_flops_per_epoch(model: Any, dataset: Any) -> int:
    """
    Calculates and returns the total FLOPs per epoch for the given model and dataset.

    Args:
        model (Any): The model for which FLOPs per epoch are to be calculated.
        dataset (Any): The dataset used for training the model.

    Returns:
        int: The total FLOPs per epoch.
    """
    global FLOPS_PER_EPOCH_COUNTER

    x, _ = dataset[0]
    flops = FlopCountAnalysis(model, x)
    total_flops = flops.total() * len(dataset)
    FLOPS_PER_EPOCH_COUNTER += total_flops
    return FLOPS_PER_EPOCH_COUNTER

def get_flops_per_batch(model: Any, batch: Any) -> int:
    """
    Calculates and returns the total FLOPs per batch for the given model and batch of data.

    Args:
        model (Any): The model for which FLOPs per batch are to be calculated.
        batch (Any): A batch of data used for inference with the model.

    Returns:
        int: The total FLOPs per batch.
    """
    global FLOPS_PER_BATCH_COUNTER
    x, _ = batch
    flops = FlopCountAnalysis(model, x)
    FLOPS_PER_BATCH_COUNTER += flops.total()
    return FLOPS_PER_BATCH_COUNTER



import zarr
import numcodecs
from enum import Enum

class CompressorType(Enum):
    NONE = None
    ZIP = "zip"
    BLOSC_ZSTD = "blosc_zstd"
    LZ4 = "l4"

COMPRESSORS_FOR_ZARR = [CompressorType.BLOSC_ZSTD, CompressorType.LZ4]

def compressor_to_type(comp): 
    if comp == CompressorType.NONE: 
        return None
    elif comp == CompressorType.ZIP: 
        return "zip"
    elif comp == CompressorType.BLOSC_ZSTD: 
        return zarr.codecs.BloscCodec(cname='zstd')
    elif comp == CompressorType.LZ4: 
        return numcodecs.lz4.LZ4()
    else: 
        raise AttributeError(f">compressor_to_type(): compressor {comp} not found")



# KEEP THESE FOR NAN eval() #
import numpy as np
from torch import nan
#############################

import pandas as pd
from yprov4ml.utils.time_utils import timestamp_to_seconds
from typing import Optional

def get_metrics(data, keyword=None):
    ms = data["entity"].keys()
    if keyword is None:
        return ms
    else:
        return [m for m in ms if keyword in m]

def get_metric(data, metric, time_in_sec=False, time_incremental=False, sort_by=None, start_at=None, end_at=None):

    if metric not in data["entity"].keys(): 
        raise AttributeError(f">get_metric({metric}) not found in prov file")

    epochs = eval(data["entity"][metric]["prov-ml:metric_epoch_list"])
    values = eval(data["entity"][metric]["prov-ml:metric_value_list"])
    times = eval(data["entity"][metric]["prov-ml:metric_timestamp_list"])

    start_at = 0 if start_at is None else start_at
    end_at = len(epochs) if end_at is None else end_at        

    epochs = epochs[start_at:end_at]
    values = values[start_at:end_at]
    times = times[start_at:end_at]
    
    # convert to minutes and sort
    if time_in_sec:
        times = [timestamp_to_seconds(ts) for ts in times]
        
    df = pd.DataFrame({"epoch": epochs, "value": values, "time": times})#.drop_duplicates()
    if time_incremental: 
        df["time"] = df["time"].diff().fillna(0)

    if sort_by is not None: 
        df = df.sort_values(by=sort_by)
    
    return df

def get_avg_metric(data, metric):
    values = eval(data["entity"][metric]["prov-ml:metric_value_list"])
    return sum(values) / len(values)

def get_sum_metric(data, metric):
    values = eval(data["entity"][metric]["prov-ml:metric_value_list"])
    return sum(values)

def get_metric_time(data, metric, time_in_sec=False): 
    times = eval(data["entity"][metric]["prov-ml:metric_timestamp_list"])
    if time_in_sec:
        times = [timestamp_to_seconds(ts) for ts in times]
    return max(times) - min(times)

def get_params(data, param):
    return [data["entity"][ent]["prov-ml:parameter_value"] for ent in data["entity"].keys() if param in ent]

def get_param(data, param): 
    if param in data["entity"].keys():
        return data["entity"][param]["prov-ml:parameter_value"]
    return None


def get_experiemnt_id(data) -> Optional[str]:
    return list(data["activity"].keys())[0]

def get_execution_command(data) -> Optional[str]:
    return get_param(data, "prov-ml:execution_command")

def get_source_code(data) -> Optional[str]: 
    return get_param(data, "prov-ml:source_code")

def get_inputs(data): 
    return get_params(data, "prov-ml:input")

def get_outputs(data): 
    return get_params(data, "prov-ml:output")




from enum import Enum

class MetricsType(Enum):
    """Enumeration for different types of metrics storage formats.

    Attributes:
        TXT (str): Represents text file format.
        ZARR (str): Represents Zarr file format.
        NETCDF (str): Represents NetCDF file format.
    """
    CSV = "csv"
    ZARR = 'zarr'
    NETCDF = 'netcdf'

def get_file_type(metrics_file_type): 
    if metrics_file_type == MetricsType.CSV:
        return ".csv"
    elif metrics_file_type == MetricsType.NETCDF: 
        return ".nc"
    elif metrics_file_type == MetricsType.ZARR: 
        return ".zarr"
    else: 
        raise NotImplementedError() 



from .logging_aux import *
from .yprov4ml import *
from .datamodel.context import Context
from .loggers.prov4ml_logger import ProvMLLogger
from .loggers.prov4ml_itwinai_logger import ProvMLItwinAILogger, LoggingItemKind




def create_activity(provDoc, record):
    if len(provDoc.get_record(record))>0:
        raise Exception(f">create_activity({record}): a record already exists")
    return provDoc.activity(record)
    
def get_activity(provDoc,record):
    records = provDoc.get_record(record)
    if len(records)>0:
        return records[0]
    else:
        raise Exception(f">get_activity({record}): a record was not found") 



import os
import argparse
from prov.model import ProvDocument
from typing import Optional

from yprov4ml.utils.file_utils import custom_prov_to_dot

def main(prov_file : str, out_file : Optional[str]): 
    if not prov_file.endswith(".json"): 
        prov_file += ".json"
    if out_file is None:
        out_file = prov_file.replace(".json", ".dot") 
    if not out_file.endswith(".dot"): 
        out_file += ".dot"

    doc = ProvDocument()
    with open(prov_file, 'r') as f:
        doc = ProvDocument.deserialize(f)

    path_dot = os.path.join(".", out_file)
    with open(path_dot, 'w') as prov_dot:
        prov_dot.write(custom_prov_to_dot(doc).to_string())
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Convert a PROV-JSON file to a DOT file')
    parser.add_argument('--prov_json', type=str, help='The PROV-JSON file to convert')
    parser.add_argument('--output', type=str, help='The output DOT file', default=None)
    args = parser.parse_args()
    
    main(args.prov_json, args.output)


import os
from typing import Optional, Union

from yprov4ml.constants import PROV4ML_DATA
from yprov4ml.utils import energy_utils
from yprov4ml.utils import flops_utils
from yprov4ml.logging_aux import log_execution_start_time, log_execution_end_time
from yprov4ml.provenance.provenance_graph import create_prov_document, create_rocrate_in_dir
from yprov4ml.utils.file_utils import save_prov_file
from yprov4ml.datamodel.metric_type import MetricsType
from yprov4ml.datamodel.compressor_type import CompressorType

def start_run(
        experiment_name: str,
        prov_user_namespace: Optional[str] = None,
        provenance_save_dir: Optional[str] = None,
        collect_all_processes: Optional[bool] = False,
        save_after_n_logs: Optional[int] = 100,
        rank : Optional[int] = None, 
        disable_codecarbon : Optional[bool] = False,
        metrics_file_type: MetricsType = MetricsType.CSV,
        csv_separator : str = ",", 
        use_compressor: Optional[Union[CompressorType, bool]] = None,
    ) -> None:
    """
    Initializes the provenance data collection and sets up various utilities for tracking.

    Parameters:
    -----------
    prov_user_namespace : str
        The user namespace to be used for organizing provenance data.
    experiment_name : Optional[str], optional
        The name of the experiment. If not provided, defaults to None.
    provenance_save_dir : Optional[str], optional
        The directory path where provenance data will be saved. If not provided, defaults to None.
    collect_all_processes : Optional[bool], optional
        Whether to collect data from all processes. Default is False.
    save_after_n_logs : Optional[int], optional
        The number of logs after which to save metrics. Default is 100.
    rank : Optional[int], optional
        The rank of the current process in a distributed setting. If not provided, defaults to None.

    Returns:
    --------
    None
    """
    PROV4ML_DATA.start_run(
        experiment_name=experiment_name, 
        prov_save_path=provenance_save_dir, 
        user_namespace=prov_user_namespace, 
        collect_all_processes=collect_all_processes, 
        save_after_n_logs=save_after_n_logs, 
        rank=rank, 
        metrics_file_type=metrics_file_type,
        csv_separator=csv_separator,
        use_compressor=use_compressor,
    )

    if not disable_codecarbon: 
        energy_utils._carbon_init()
    flops_utils._init_flops_counters()

    log_execution_start_time()

def end_run(
        create_graph: Optional[bool] = False, 
        create_svg: Optional[bool] = False, 
        crate_ro_crate: Optional[bool]=False,
    ):  
    """
    Finalizes the provenance data collection and optionally creates visualization and provenance collection files.

    Parameters:
    -----------
    create_graph : Optional[bool], optional
        Whether to create a graph representation of the provenance data. Default is False.
    create_svg : Optional[bool], optional
        Whether to create an SVG file for the graph visualization. Default is False. 
        Must be set to True only if `create_graph` is also True.
    create_provenance_collection : Optional[bool], optional
        Whether to create a collection of provenance data from all runs. Default is False.

    Raises:
    -------
    ValueError
        If `create_svg` is True but `create_graph` is False.

    Returns:
    --------
    None
    """

    if not PROV4ML_DATA.is_collecting: return
    
    log_execution_end_time()

    found = False
    for root, _, filenames in os.walk('.'):
        for filename in filenames:
            if filename == "requirements.txt": 
                PROV4ML_DATA.add_artifact("requirements", os.path.join(root, filename), step=0, context=None, is_input=True)
                found = True
            if found: break
        if found: break

    PROV4ML_DATA.save_all_metrics()

    doc = create_prov_document()
   
    graph_filename = f'prov_{PROV4ML_DATA.PROV_JSON_NAME}.json'
    
    if not os.path.exists(PROV4ML_DATA.EXPERIMENT_DIR):
        os.makedirs(PROV4ML_DATA.EXPERIMENT_DIR, exist_ok=True)
    
    path_graph = os.path.join(PROV4ML_DATA.EXPERIMENT_DIR, graph_filename)
    save_prov_file(doc, path_graph, create_graph, create_svg)

    if crate_ro_crate: 
        create_rocrate_in_dir(PROV4ML_DATA.EXPERIMENT_DIR)




from aenum import Enum

class Context(Enum): 
    # EXPERIMENT = "EXPERIMENT"
    TRAINING = "TRAINING"
    VALIDATION = "VALIDATION"
    TESTING = "TESTING"
    DATASETS = "DATASETS"
    MODELS = "MODELS"

    @staticmethod
    def get_context_from_string(context: str): 
        """
        Returns the context enum from a string.

        Parameters:
            context (str): The context string.

        Returns:
            Context: The context enum.
        """
        try: 
            context = eval(context)
            if type(context) == Context: 
                return context
            else: 
                raise ValueError(f"Invalid context: {context}")
        except: 
            if context == 'training' or context == 'Context.TRAINING':
                return Context.TRAINING
            elif context == 'testing' or context == 'Context.TESTING':
                return Context.TESTING
            elif context == 'validation' or context == 'Context.VALIDATION':
                return Context.VALIDATION
            elif context == 'models' or context == 'Context.MODELS':
                return Context.MODELS
            elif context == 'datasets' or context == 'Context.DATASETS':
                return Context.DATASETS
            else:
                raise ValueError(f"Not a context: {context}")
                    



import time
import pandas as pd

DATA = {}

def get_time() -> float:
    """
    Returns the current time in seconds since the epoch (Unix timestamp).

    Returns:
        float: The current time in seconds since the epoch.
    """
    return time.time()

def timestamp_to_minutes(ts):
    return ts / 60000

def timestamp_to_seconds(ts):
    return ts / 1000


def check_timer(label): 
    global DATA
    t = time.time()

    if label in DATA.keys(): 
        DATA[label] = t - DATA[label]
    else: 
        DATA[label] = t

def save_times(filename): 
    global DATA 
    pd.DataFrame(DATA, index=[0]).T.to_csv(filename)


import os
import torch
import json
import warnings
import prov.model as prov

from torch.utils.data import DataLoader, Subset, Dataset, RandomSampler
from typing import Any, Optional, Union

from yprov4ml.datamodel.attribute_type import LoggingItemKind
from yprov4ml.utils import energy_utils, flops_utils, system_utils, time_utils, funcs
from yprov4ml.datamodel.context import Context
from yprov4ml.constants import PROV4ML_DATA, VERBOSE
    
def log_metric(
        key: str, 
        value: float, 
        context: Optional[Context] = None, 
        step: Optional[int] = 0, 
        source: Optional[LoggingItemKind] = None, 
    ) -> None:
    """
    Logs a metric with the specified key, value, and context.

    Args:
        key (str): The key of the metric.
        value (float): The value of the metric.
        context (Context): The context in which the metric is recorded.
        step (Optional[int], optional): The step number for the metric. Defaults to None.
        source (LoggingItemKind, optional): The source of the logging item. Defaults to None.

    Returns:
        None
    """
    PROV4ML_DATA.add_metric(key, value, step, context=context, source=source)

def log_execution_start_time() -> None:
    """Logs the start time of the current execution. """
    return log_param("execution_start_time", time_utils.get_time())

def log_execution_end_time() -> None:
    """Logs the end time of the current execution."""
    return log_param("execution_end_time", time_utils.get_time())

def log_current_execution_time(label: str, context: Context, step: Optional[int] = None) -> None:
    """Logs the current execution time under the given label.
    
    Args:
        label (str): The label to associate with the logged execution time.
        context (mlflow.tracking.Context): The MLflow tracking context.
        step (Optional[int], optional): The step number for the logged execution time. Defaults to None.

    Returns:
        None
    """
    return log_metric(label, time_utils.get_time(), context, step=step, source=LoggingItemKind.EXECUTION_TIME)

def log_param(key: str, value: Any, context : Context = None) -> None:
    """Logs a single parameter key-value pair. 
    
    Args:
        key (str): The key of the parameter.
        value (Any): The value of the parameter.

    Returns:
        None
    """
    PROV4ML_DATA.add_parameter(key,value, context)

def _get_model_memory_footprint(model_name: str, model: Union[torch.nn.Module, Any]) -> dict:
    """Logs the memory footprint of the provided model.
    
    Args:
        model (Union[torch.nn.Module, Any]): The model whose memory footprint is to be logged.
        model_name (str, optional): Name of the model. Defaults to "default".

    Returns:
        None
    """
    ret = {"model_name": model_name}


    total_params = sum(p.numel() for p in model.parameters())
    try: 
        if hasattr(model, "trainer"): 
            precision_to_bits = {"64": 64, "32": 32, "16": 16, "bf16": 16}
            if hasattr(model.trainer, "precision"):
                precision = precision_to_bits.get(model.trainer.precision, 32)
            else: 
                precision = 32
        else: 
            precision = 32
    except RuntimeError: 
        if VERBOSE: 
            warnings.warn("Could not determine precision, defaulting to 32 bits. Please make sure to provide a model with a trainer attached, this is often due to calling this before the trainer.fit() method")
        precision = 32
    
    precision_megabytes = precision / 8 / 1e6

    memory_per_model = total_params * precision_megabytes
    memory_per_grad = total_params * 4 * 1e-6
    memory_per_optim = total_params * 4 * 1e-6
    
    ret[f"{PROV4ML_DATA.PROV_PREFIX}:total_params"] = total_params
    ret[f"{PROV4ML_DATA.PROV_PREFIX}:memory_of_model"] = memory_per_model
    ret[f"{PROV4ML_DATA.PROV_PREFIX}:total_memory_load_of_model"] = memory_per_model + memory_per_grad + memory_per_optim

    return ret

def _get_nested_model_desc(m: torch.nn.Module):
    children = dict(m.named_children())
    output = {}
    if children == {}:
        node = {}
        funcs._safe_get_model_attr(node, m, "type", attr_label="layer_type")
        funcs._safe_get_model_attr(node, m, "in_features")
        funcs._safe_get_model_attr(node, m, "out_features")
        funcs._safe_get_model_attr(node, m, "in_channels")
        funcs._safe_get_model_attr(node, m, "out_channels")
        funcs._safe_get_model_attr(node, m, "kernel_size")
        funcs._safe_get_model_attr(node, m, "stride")
        funcs._safe_get_model_attr(node, m, "padding")
        funcs._safe_get_model_attr(node, m, "weight.dtype", attr_label="dtype")
        # safe_get_attr(node, m, "bias", attr_label="layer_bias")
        return node
    else:
        for name, child in children.items():
            output[f"{PROV4ML_DATA.PROV_PREFIX}:{name}"] = _get_nested_model_desc(child)

    return output

def _get_model_layers_description(model_name : str, model: Union[torch.nn.Module, Any]) -> prov.ProvEntity: 
    mo = _get_nested_model_desc(model)
    
    path = os.path.join(PROV4ML_DATA.ARTIFACTS_DIR, model_name)
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)
    path = os.path.join(path, f"{model_name}_layers_description.json")
    with open(path, "w") as fp:
        json.dump(mo , fp) 

    return {f"{PROV4ML_DATA.PROV_PREFIX}:layers_description_path": path}

def log_model(
        model_name: str, 
        model: Union[torch.nn.Module, Any], 
        context : Optional[Context] = Context.MODELS, 
        log_model_info: bool = True, 
        log_model_layers : bool = False,
        is_input: bool = False,
    ) -> None:
    """Logs the provided model as artifact and logs memory footprint of the model. 
    
    Args:
        model (Union[torch.nn.Module, Any]): The model to be logged.
        model_name (str, optional): Name of the model. Defaults to "default".
        log_model_info (bool, optional): Whether to log model memory footprint. Defaults to True.
        log_model_layers (bool, optional): Whether to log model layers details. Defaults to False.
        log_as_artifact (bool, optional): Whether to log the model as an artifact. Defaults to True.
    """
    e = save_model_version(model_name, model, context, incremental=False, is_input=is_input)

    if log_model_info:
        d = _get_model_memory_footprint(model_name, model)
        e.add_attributes(d)

    if log_model_layers: 
        d = _get_model_layers_description(model_name, model)
        e.add_attributes(d)

     
def log_flops_per_epoch(label: str, model: Any, dataset: Any, context: Context, step: Optional[int] = None) -> None:
    """Logs the number of FLOPs (floating point operations) per epoch for the given model and dataset.
    
    Args:
        label (str): The label to associate with the logged FLOPs per epoch.
        model (Any): The model for which FLOPs per epoch are to be logged.
        dataset (Any): The dataset used for training the model.
        context (mlflow.tracking.Context): The MLflow tracking context.
        step (Optional[int], optional): The step number for the logged FLOPs per epoch. Defaults to None.

    Returns:
        None
    """
    return log_metric(label, flops_utils.get_flops_per_epoch(model, dataset), context, step=step, source=LoggingItemKind.FLOPS_PER_EPOCH)

def log_flops_per_batch(label: str, model: Any, batch: Any, context: Context, step: Optional[int] = None) -> None:
    """Logs the number of FLOPs (floating point operations) per batch for the given model and batch of data.
    
    Args:
        label (str): The label to associate with the logged FLOPs per batch.
        model (Any): The model for which FLOPs per batch are to be logged.
        batch (Any): A batch of data used for inference with the model.
        context (mlflow.tracking.Context): The MLflow tracking context.
        step (Optional[int], optional): The step number for the logged FLOPs per batch. Defaults to None.

    Returns:
        None
    """
    return log_metric(label, flops_utils.get_flops_per_batch(model, batch), context, step=step, source=LoggingItemKind.FLOPS_PER_BATCH)

def log_system_metrics(
    context: Context,
    step: Optional[int] = None,
    ) -> None:
    """Logs system metrics such as CPU usage, memory usage, disk usage, and GPU metrics.

    Args:
        context (mlflow.tracking.Context): The MLflow tracking context.
        step (Optional[int], optional): The step number for the logged metrics. Defaults to None.

    Returns:
        None
    """
    log_metric("cpu_usage", system_utils.get_cpu_usage(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)
    log_metric("memory_usage", system_utils.get_memory_usage(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)
    log_metric("disk_usage", system_utils.get_disk_usage(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)

    
    log_metric("gpu_memory_usage", system_utils.get_gpu_memory_usage(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)
    log_metric("gpu_usage", system_utils.get_gpu_usage(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)
    log_metric("gpu_temperature", system_utils.get_gpu_temperature(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)
    log_metric("gpu_power_usage", system_utils.get_gpu_power_usage(), context, step=step, source=LoggingItemKind.SYSTEM_METRIC)

def log_carbon_metrics(
    context: Context,
    step: Optional[int] = None,
    ):
    """Logs carbon emissions metrics such as energy consumed, emissions rate, and power consumption.
    
    Args:
        context (mlflow.tracking.Context): The MLflow tracking context.
        step (Optional[int], optional): The step number for the logged metrics. Defaults to None.
    
    Returns:
        None
    """    
    emissions = energy_utils.stop_carbon_tracked_block()
   
    log_metric("emissions", emissions.energy_consumed, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("emissions_rate", emissions.emissions_rate, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("cpu_power", emissions.cpu_power, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("gpu_power", emissions.gpu_power, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("ram_power", emissions.ram_power, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("cpu_energy", emissions.cpu_energy, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("gpu_energy", emissions.gpu_energy, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("ram_energy", emissions.ram_energy, context, step=step, source=LoggingItemKind.CARBON_METRIC)
    log_metric("energy_consumed", emissions.energy_consumed, context, step=step, source=LoggingItemKind.CARBON_METRIC)

def log_artifact(
        artifact_name : str, 
        artifact_path : str, 
        context: Optional[Context] = None,
        step: Optional[int] = None, 
        log_copy_in_prov_directory : bool = True, 
        is_model : bool = False, 
        is_input : bool = False, 
    ) -> prov.ProvEntity:
    """
    Logs the specified artifact to the given context.

    Parameters:
        artifact_path (str): The file path of the artifact to log.
        context (Context): The context in which the artifact is logged.
        step (Optional[int]): The step or epoch number associated with the artifact. Defaults to None.
        timestamp (Optional[int]): The timestamp associated with the artifact. Defaults to None.

    Returns:
        None
    """
    return PROV4ML_DATA.add_artifact(
        artifact_name=artifact_name, 
        artifact_path=artifact_path, 
        step=step, 
        context=context, 
        log_copy_in_prov_directory=log_copy_in_prov_directory, 
        is_model=is_model, 
        is_input=is_input, 
    )

def save_model_version(
        model_name: str, 
        model: Union[torch.nn.Module, Any], 
        context: Optional[Context] = None, 
        step: Optional[int] = None, 
        incremental : bool = True, 
        is_input : bool =False, 
    ) -> prov.ProvEntity:
    """
    Saves the state dictionary of the provided model and logs it as an artifact.
    
    Parameters:
        model (torch.nn.Module): The PyTorch model to be saved.
        model_name (str): The name under which to save the model.
        context (Context): The context in which the model is saved.
        step (Optional[int]): The step or epoch number associated with the saved model. Defaults to None.
        timestamp (Optional[int]): The timestamp associated with the saved model. Defaults to None.

    Returns:
        None
    """

    path = os.path.join(PROV4ML_DATA.ARTIFACTS_DIR, model_name)
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)

    # count all models with the same name stored at "path"
    if incremental: 
        num_files = len([file for file in os.listdir(path) if str(file).startswith(model_name)])
        fn = os.path.join(path, f"{model_name}_{num_files}.pt")
        torch.save(model.state_dict(), fn)
        return log_artifact(f"{model_name}_{num_files}", fn, context=context, step=step, log_copy_in_prov_directory=False, is_model=True, is_input=is_input)
    else: 
        fn = os.path.join(path, f"{model_name}.pt")
        torch.save(model.state_dict(), fn)
        return log_artifact(model_name, fn, context=context, step=step, log_copy_in_prov_directory=False, is_model=True, is_input=is_input)

def log_dataset(
        dataset_label : str, 
        dataset : Union[DataLoader, Subset, Dataset], 
        context : Optional[Context] = Context.DATASETS, 
        log_dataset_info : bool = True, 
        ): 
    """
    Logs dataset statistics such as total samples and total steps.

    Args:
        dataset (Union[DataLoader, Subset, Dataset]): The dataset for which statistics are to be logged.
        label (str): The label to associate with the logged dataset statistics.

    Returns:
        None
    """

    e = log_artifact(f"{dataset_label}", "", context=context, log_copy_in_prov_directory=False, is_model=False, is_input=True)
    
    if not log_dataset_info: return

    e.add_attributes({f"{PROV4ML_DATA.PROV_PREFIX}:{dataset_label}_stat_total_samples": len(dataset)})
    # handle datasets from DataLoader
    if isinstance(dataset, DataLoader):
        dl = dataset
        dataset = dl.dataset
        attrs = {
            f"{PROV4ML_DATA.PROV_PREFIX}:{dataset_label}_stat_batch_size": dl.batch_size, 
            f"{PROV4ML_DATA.PROV_PREFIX}:{dataset_label}_stat_num_workers": dl.num_workers, 
            f"{PROV4ML_DATA.PROV_PREFIX}:{dataset_label}_stat_shuffle": isinstance(dl.sampler, RandomSampler), 
            f"{PROV4ML_DATA.PROV_PREFIX}:{dataset_label}_stat_total_steps": len(dl), 
        }
        e.add_attributes(attrs)

    elif isinstance(dataset, Subset):
        dl = dataset
        dataset = dl.dataset
        e.add_attributes({f"{dataset_label}_stat_total_steps": len(dl)})

def log_execution_command(cmd: str, path : str) -> None:
    """
    Logs the execution command.
    
    Args:
        cmd (str): The command to be logged.
    """
    path = os.path.join("", "workspace", f"{PROV4ML_DATA.CLEAN_EXPERIMENT_NAME}_{PROV4ML_DATA.RUN_ID}", "artifacts", path)
    log_param("execution_command", cmd + " " + path)

def log_source_code(path: Optional[str] = None) -> None:
    """
    Logs the source code location, either from a Git repository or a specified path.
    
    Args: 
        path (Optional[str]): The path to the source code. If None, attempts to retrieve from Git.
    """
    PROV4ML_DATA.add_source_code(path)

def create_context(context : str, is_subcontext_of=None): 
    PROV4ML_DATA.add_context(context, is_subcontext_of=is_subcontext_of)



import os
import prov.model as prov

from yprov4ml.constants import PROV4ML_DATA

def save_prov_file(
        doc : prov.ProvDocument,
        prov_file : str,
        create_graph : bool =False, 
        create_svg : bool =False
    ) -> None:
    """
    Save the provenance document to a file.

    Parameters:
    -----------
    doc : prov.ProvDocument
        The provenance document to save.
    prov_file : str
        The path to the file where the provenance document will be saved.
    create_graph : bool 
        A flag to indicate if a graph should be created. Defaults to False.
    create_svg : bool
        A flag to indicate if an SVG should be created. Defaults to False.
    
    Returns:
        None
    """

    with open(prov_file, 'w') as prov_graph:
        doc.serialize(prov_graph)

    if create_svg and not create_graph:
        raise ValueError("Cannot create SVG without creating the graph.")

    if create_graph:
        dot_filename = os.path.basename(prov_file).replace(".json", ".dot")
        path_dot = os.path.join(PROV4ML_DATA.EXPERIMENT_DIR, dot_filename)
        with open(path_dot, 'w') as prov_dot:
            prov_dot.write(custom_prov_to_dot(doc).to_string())

    if create_svg:
        svg_filename = os.path.basename(prov_file).replace(".json", ".svg")
        path_svg = os.path.join(PROV4ML_DATA.EXPERIMENT_DIR, svg_filename)
        os.system(f"dot -Tsvg {path_dot} > {path_svg}")


from prov.model import (
    ProvException,
    Identifier,
    PROV_ATTRIBUTE_QNAMES,
    datetime,
)
from prov.graph import INFERRED_ELEMENT_CLASS
from prov.dot import ANNOTATION_STYLE, ANNOTATION_LINK_STYLE, ANNOTATION_START_ROW, ANNOTATION_ROW_TEMPLATE, ANNOTATION_END_ROW, DOT_PROV_STYLE, GENERIC_NODE_STYLE, sorted_attributes
from xml.sax.saxutils import escape
from datetime import datetime
import pydot

def custom_prov_to_dot(
    bundle,
    show_nary=True,
    use_labels=False,
    direction="RL",
    show_element_attributes=True,
    show_relation_attributes=True,
):
    """
    Convert a provenance bundle/document into a DOT graphical representation.

    :param bundle: The provenance bundle/document to be converted.
    :type bundle: :class:`ProvBundle`
    :param show_nary: shows all elements in n-ary relations.
    :type show_nary: bool
    :param use_labels: uses the prov:label property of an element as its name (instead of its identifier).
    :type use_labels: bool
    :param direction: specifies the direction of the graph. Valid values are "BT" (default), "TB", "LR", "RL".
    :param show_element_attributes: shows attributes of elements.
    :type show_element_attributes: bool
    :param show_relation_attributes: shows attributes of relations.
    :type show_relation_attributes: bool
    :returns:  :class:`pydot.Dot` -- the Dot object.
    """
    if direction not in {"BT", "TB", "LR", "RL"}:
        # Invalid direction is provided
        direction = "BT"  # reset it to the default value
    maindot = pydot.Dot(graph_type="digraph", rankdir=direction, charset="utf-8")

    node_map = {}
    count = [0, 0, 0, 0]  # counters for node ids

    def _bundle_to_dot(dot, bundle):
        def _attach_attribute_annotation(node, record):
            # Adding a node to show all attributes
            attributes = []
                # Truncate the value if it is too long
            for attr_name, value in record.attributes: 
                if attr_name not in PROV_ATTRIBUTE_QNAMES:
                    try: 
                        if len(value) > 100: 
                            attributes.append((attr_name, str(value[:100]) + "..."))
                        else: 
                            attributes.append((attr_name, value))
                    except TypeError: 
                        attributes.append((attr_name, value))

            if not attributes or len(attributes) == 0:
                return  # No attribute to display

            # Sort the attributes.
            attributes = sorted_attributes(record.get_type(), attributes)

            ann_rows = [ANNOTATION_START_ROW]
            ann_rows.extend(
                ANNOTATION_ROW_TEMPLATE
                % (
                    attr.uri,
                    escape(str(attr)),
                    ' href="%s"' % value.uri if isinstance(value, Identifier) else "",
                    escape(
                        str(value)
                        if not isinstance(value, datetime)
                        else str(value.isoformat())
                    ),
                )
                for attr, value in attributes
            )
            ann_rows.append(ANNOTATION_END_ROW)
            count[3] += 1
            annotations = pydot.Node(
                "ann%d" % count[3], label="\n".join(ann_rows), **ANNOTATION_STYLE
            )
            dot.add_node(annotations)
            dot.add_edge(pydot.Edge(annotations, node, **ANNOTATION_LINK_STYLE))

        def _add_bundle(bundle):
            count[2] += 1
            subdot = pydot.Cluster(
                graph_name="c%d" % count[2], URL=f'"{bundle.identifier.uri}"'
            )
            if use_labels:
                if bundle.label == bundle.identifier:
                    bundle_label = f'"{bundle.label}"'
                else:
                    # Fancier label if both are different. The label will be the main
                    # node text, whereas the identifier will be a kind of subtitle.
                    bundle_label = (
                        f"<{bundle.label}<br />"
                        f'<font color="#333333" point-size="10">'
                        f'{bundle.identifier}</font>>'
                    )
                subdot.set_label(f'"{bundle_label}"')
            else:
                subdot.set_label('"%s"' % str(bundle.identifier))
            _bundle_to_dot(subdot, bundle)
            dot.add_subgraph(subdot)
            return subdot

        def _add_node(record):
            count[0] += 1
            node_id = "n%d" % count[0]
            if use_labels:
                if record.label == record.identifier:
                    node_label = f'"{record.label}"'
                else:
                    # Fancier label if both are different. The label will be
                    # the main node text, whereas the identifier will be a
                    # kind of subtitle.
                    node_label = (
                        f"<{record.label}<br />"
                        f'<font color="#333333" point-size="10">'
                        f'{record.identifier}</font>>'
                    )
            else:
                node_label = f'"{record.identifier}"'

            uri = record.identifier.uri
            style = DOT_PROV_STYLE[record.get_type()]
            node = pydot.Node(node_id, label=node_label, URL='"%s"' % uri, **style)
            node_map[uri] = node
            dot.add_node(node)

            if show_element_attributes:
                _attach_attribute_annotation(node, rec)
            return node

        def _add_generic_node(qname, prov_type=None):
            count[0] += 1
            node_id = "n%d" % count[0]
            node_label = f'"{qname}"'

            uri = qname.uri
            style = GENERIC_NODE_STYLE[prov_type] if prov_type else DOT_PROV_STYLE[0]
            node = pydot.Node(node_id, label=node_label, URL='"%s"' % uri, **style)
            node_map[uri] = node
            dot.add_node(node)
            return node

        def _get_bnode():
            count[1] += 1
            bnode_id = "b%d" % count[1]
            bnode = pydot.Node(bnode_id, label='""', shape="point", color="gray")
            dot.add_node(bnode)
            return bnode

        def _get_node(qname, prov_type=None):
            if qname is None:
                return _get_bnode()
            uri = qname.uri
            if uri not in node_map:
                _add_generic_node(qname, prov_type)
            return node_map[uri]

        records = bundle.get_records()
        relations = []
        for rec in records:
            if rec.is_element():
                _add_node(rec)
            else:
                # Saving the relations for later processing
                relations.append(rec)

        if not bundle.is_bundle():
            for bundle in bundle.bundles:
                _add_bundle(bundle)

        for rec in relations:
            args = rec.args
            # skipping empty records
            if not args:
                continue
            # picking element nodes
            attr_names, nodes = zip(
                *(
                    (attr_name, value)
                    for attr_name, value in rec.formal_attributes
                    if attr_name in PROV_ATTRIBUTE_QNAMES
                )
            )
            inferred_types = list(map(INFERRED_ELEMENT_CLASS.get, attr_names))
            other_attributes = [
                (attr_name, value)
                for attr_name, value in rec.attributes
                if attr_name not in PROV_ATTRIBUTE_QNAMES
            ]
            add_attribute_annotation = show_relation_attributes and other_attributes
            add_nary_elements = len(nodes) > 2 and show_nary
            style = DOT_PROV_STYLE[rec.get_type()]
            if len(nodes) < 2:  # too few elements for a relation?
                continue  # cannot draw this

            if add_nary_elements or add_attribute_annotation:
                # a blank node for n-ary relations or the attribute annotation
                bnode = _get_bnode()

                # the first segment
                dot.add_edge(
                    pydot.Edge(
                        _get_node(nodes[0], inferred_types[0]),
                        bnode,
                        arrowhead="none",
                        **style,
                    )
                )
                style = dict(style)  # copy the style
                del style["label"]  # not showing label in the second segment
                # the second segment
                dot.add_edge(
                    pydot.Edge(bnode, _get_node(nodes[1], inferred_types[1]), **style)
                )
                if add_nary_elements:
                    style["color"] = "gray"  # all remaining segment to be gray
                    style["fontcolor"] = "dimgray"  # text in darker gray
                    for attr_name, node, inferred_type in zip(
                        attr_names[2:], nodes[2:], inferred_types[2:]
                    ):
                        if node is not None:
                            style["label"] = attr_name.localpart
                            dot.add_edge(
                                pydot.Edge(
                                    bnode, _get_node(node, inferred_type), **style
                                )
                            )
                if add_attribute_annotation:
                    _attach_attribute_annotation(bnode, rec)
            else:
                # show a simple binary relations with no annotation
                dot.add_edge(
                    pydot.Edge(
                        _get_node(nodes[0], inferred_types[0]),
                        _get_node(nodes[1], inferred_types[1]),
                        **style,
                    )
                )

    try:
        unified = bundle.unified()
    except ProvException:
        # Could not unify this bundle
        # try the original document anyway
        unified = bundle

    _bundle_to_dot(maindot, unified)
    return maindot




from typing import Any, Optional

from yprov4ml.utils.funcs import get_current_time_millis

class ArtifactInfo:
    """
    Holds information about an artifact, including its path, value, step, context, and timestamps.

    Attributes:
        path (str): The path of the artifact.
        value (Any): The value of the artifact.
        step (Optional[int]): The step number associated with the artifact.
        context (Optional[Any]): The context of the artifact.
        creation_timestamp (Optional[int]): The creation timestamp of the artifact.
        last_modified_timestamp (Optional[int]): The last modified timestamp of the artifact.
        is_model_version (bool): Indicates if the artifact is a PyTorch model.
    """
    def __init__(
        self, 
        name: str, 
        value: Any = None, 
        step: Optional[int] = None, 
        context: Optional[Any] = None, 
        is_model : bool = False, 
    ) -> None:
        self.path = name
        self.value = value
        self.step = step
        self.context = context
        self.creation_timestamp = get_current_time_millis()
        self.last_modified_timestamp = get_current_time_millis()

        self.is_model_version = is_model

    def update(
        self, 
        value: Any = None, 
        step: Optional[int] = None, 
        context: Optional[Any] = None
    ) -> None:
        """
        Updates the artifact information with the provided values.

        Parameters:
            value (Any): The new value of the artifact. Defaults to None.
            step (Optional[int]): The new step number for the artifact. Defaults to None.
            context (Optional[Any]): The new context of the artifact. Defaults to None.
        """
        self.value = value if value is not None else self.value
        self.step = step if step is not None else self.step
        self.context = context if context is not None else self.context
        self.last_modified_timestamp = get_current_time_millis()




from torch.utils.data import Dataset

class IndexedDatasetWrapper(Dataset):
    """
    Modifies the given Dataset class to return a tuple data, target, index
    instead of just data, target.
    """

    def __init__(self, dataset):
        super().__init__()

        self.dataset = dataset

    def __len__(self): 
        return len(self.dataset)
    
    def __getitem__(self, index):
        batch = self.dataset[index]
        return index, batch


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='yprov4ml',
    version='1.0.0',
    packages=find_packages(),
    install_requires=required,  # Loaded from requirements.txt
    extras_require={
        'apple': [
            # Optional dependencies for Apple/Mac
            'apple_gpu==0.3.0'
        ], 
        'amd': [
            # Optional dependencies for AMD
            'amd_gpu==0.3.0', 
            'pyamdgpuinfo==2.1.6',
        ], 
        'nvidia': [
            # Optional dependencies for NVIDIA
            'nvitop==1.3.2',
        ]
    }
)
